/*
 * Copyright (C) 1998-2000 Luke Howard. All rights reserved.
 * CONFIDENTIAL
 * $Id$
 *
 * Implementation of GSS-API client side binding for SASL
 */

/* #define HACK_SERVICE_NAME */

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <winsock.h>
//#include <sys/uio.h>
#ifdef HEIMDAL
//#include <gssapi.h>
#elif SEAM
#include <gssapi/gssapi.h>
#include <gssapi/gssapi_ext.h>
#elif _WINDOWS
//#include "GssAPI.h"
#include <gssapi/gssapi.h>
#include <gssapi/gssapi_generic.h>
#else

//#include <gssapi/gssapi_generic.h>
//#include <gssapi/gssapi_krb5.h>
#endif				/* HEIMDAL */
//#include <netinet/in.h>

#ifdef linux
#define LINUX2_0
#endif

#include "ldap-int.h" //orig
//#include "ldap.h"
//#include "lber.h"
#include "lber-int.h" //orig
#include "gssldap-int.h"
#include <errno.h>
#include <stdio.h>
#include <sys/types.h>

#ifdef GSSSASL_DEBUG
static int debug_ = 1;
#define TRACE(x) do { if (debug_) fprintf(stderr, "%s\n", x); fflush(stderr); } while (0);
#define LDAP_PERROR(ld, x) do { if (debug_) ldap_perror(ld, x); } while (0);
#define LOG_STATUS(msg, min, maj) log_status(msg, min, maj)
#else
static int debug_ = 1;
#define TRACE(x) do { if (debug_) fprintf(stderr, "%s\n", x); fflush(stderr); } while (0);
#define LDAP_PERROR(ld, x) do { if (debug_) ldap_perror(ld, x); } while (0);
#define LOG_STATUS(msg, min, maj) log_status(msg, min, maj)
#endif				/* GSSSASL_DEBUG */

#ifdef HACK_SERVICE_NAME
char *__service = NULL;
#endif

/*
 * The read and write fns need to access the context in here and
 * there doesn't appear to be a way to pass this info to them.
 */
static gss_ctx_id_t security_context;
static int security_layer = 0;
static int security_token_size = 0;
/////////////////////////////added
typedef struct iovec {
        void    *iov_base;
        size_t  iov_len;
} iovec_t;
LDAP_API(int) LDAP_C LDAP_CALLBACK ldap_gssapi_read(LBER_SOCKET sock, void *data, int len);
LDAP_API(int) LDAP_C LDAP_CALLBACK ldap_gssapi_write(LBER_SOCKET sock, const void *data, int len);
////////////////////////////
#ifdef GSSSASL_DEBUG
/*
 * Dump the token to stderr
 */
static void print_token(gss_buffer_t tok)
{
    int i;
    unsigned char *p = tok->value;

    for (i = 0; i < tok->length; i++, p++) {
	fprintf(stderr, "%02x ", *p);
	if ((i % 16) == 15) {
	    fprintf(stderr, "\n");
	}
    }
    fprintf(stderr, "\n");
    fflush(stderr);
    return;
}

/*
 * Handle a GSS-API error
 */
static void log_status_impl(const char *reason,
			    OM_uint32 res,
			    int type)
{
    OM_uint32 maj_stat, min_stat;
    gss_buffer_desc msg;
    OM_uint32 msg_ctx;


    msg_ctx = 0;
    while (1) {
	maj_stat = gss_display_status(&min_stat,
				      res,
				      type,
				      GSS_C_NULL_OID,
				      &msg_ctx,
				      &msg);

	if (debug_) {
	    fprintf(stderr, "ldap_gssapi_bind: %s: %s\n",
		    (char *) msg.value,
		    reason);
	}
	(void) gss_release_buffer(&min_stat, &msg);

	if (!msg_ctx)
	    break;
    }
    return;
}

/*
 * Cover function to handle a GSS-API error
 */
static void log_status(const char *reason,
		       OM_uint32 maj_stat,
		       OM_uint32 min_stat)
{
    log_status_impl(reason, maj_stat, GSS_C_GSS_CODE);
    log_status_impl(reason, min_stat, GSS_C_MECH_CODE);
    return;
}
#endif				/* GSSSASL_DEBUG */

/*
 * Send a GSS-API token as part of a SASL BindRequest
 */
static int send_token(gssldap_client_state_t state,
		      gss_buffer_t send_tok)
{
    struct berval cred;

    TRACE("==> send_token");

    cred.bv_val = send_tok->value;
    cred.bv_len = send_tok->length;

    if (ldap_sasl_bind(state->ld,
		       state->binddn,
		       GSSAPI_SASL_NAME,
		       &cred,
		       NULL,
		       NULL,
		       &state->msgid) != LDAP_SUCCESS) {
	LDAP_PERROR(state->ld, "send_token");
	TRACE("<== send_token");
	return -1;
    }
    TRACE("<== send_token");

    return 0;
}

/*
 * Parse the final result sent back from the server.
 */
static int parse_bind_result(gssldap_client_state_t state)
{
    LDAPMessage *res = NULL, *msg = NULL;
    int rc;

    TRACE("==> parse_bind_result");

    if (ldap_result(state->ld,
		    state->msgid,
		    LDAP_MSG_ALL,
		    NULL,
		    &res) <= 0) {
	LDAP_PERROR(state->ld, "ldap_result");
	TRACE("<== parse_bind_result");
	return -1;
    }
    for (msg = ldap_first_message(state->ld, res);
	 msg != NULL;
	 msg = ldap_next_message(state->ld, msg)) {
	fprintf(stdout, "ldap_first_message \n");
	if (ldap_msgtype(msg) == LDAP_RES_BIND) {
		 fprintf(stdout, "ldap_first_message::LDAP_RES_BIND \n");

	   ldap_parse_result(state->ld, msg, &rc, NULL, NULL, NULL, NULL, 0); //VIKAS
	    break;
	}
    }

    ldap_msgfree(res);
    state->msgid = -1;

    TRACE("<== parse_bind_result");

    if (rc == LDAP_SUCCESS) {
	return 0;
    }
    state->rc = rc;
    return -1;
}

/*
 * Receive a GSS-API token from a SASL BindResponse
 * The contents of recv_tok must be freed by the
 * caller.
 */
static int recv_token(gssldap_client_state_t state,
		      gss_buffer_t recv_tok)
{
    struct berval *servercred = NULL;
    LDAPMessage *res = NULL, *msg = NULL;
    int rc;

    TRACE("==> recv_token");

    if (ldap_result(state->ld, state->msgid, LDAP_MSG_ALL, NULL, &res) <= 0) {
	LDAP_PERROR(state->ld, "ldap_result");
	TRACE("<== recv_token");
	return -1;
    }
    recv_tok->value = NULL;
    recv_tok->length = 0;

    for (msg = ldap_first_message(state->ld, res);
	 msg != NULL;
	 msg = ldap_next_message(state->ld, msg)) {
	if (ldap_msgtype(msg) == LDAP_RES_BIND && servercred == NULL) {
	    rc = ldap_parse_sasl_bind_result(state->ld,
					     msg,
					     &servercred,
					     0);
	    if (rc == LDAP_SUCCESS && servercred != NULL) {
		/* copy it so it can be freed by GSS-API */
		recv_tok->value = malloc(servercred->bv_len);
		if (recv_tok->value != NULL) {
		    memcpy(recv_tok->value, servercred->bv_val, servercred->bv_len);
		    recv_tok->length = servercred->bv_len;
		}
		break;
	    } else {
		state->rc = rc;
	    }
	}
    }

    ldap_msgfree(res);

    state->msgid = -1;

    TRACE("<== recv_token");

    if (servercred != NULL) {
		ldap_ber_free(servercred,1);//ldap_ber_free(servercred,1);//nslberi_free(servercred); //free(servercred);//changed
		TRACE("<== recv_token");
		return 0;
    }
    if (state->rc == LDAP_SUCCESS) {
	state->rc = LDAP_OPERATIONS_ERROR;
    } else {
	LDAP_PERROR(state->ld, "recv_token");
    }

    TRACE("<== recv_token");

    return -1;
}

/*
 * The client calls GSS_Init_sec_context, passing in 0 for
 * input_context_handle (initially) and a targ_name equal to output_name
 * from GSS_Import_Name called with input_name_type of
 * GSS_C_NT_HOSTBASED_SERVICE and input_name_string of
 * "service@hostname" where "service" is the service name specified in
 * the protocol's profile, and "hostname" is the fully qualified host
 * name of the server.  The client then responds with the resulting
 * output_token.  If GSS_Init_sec_context returns GSS_S_CONTINUE_NEEDED,
 * then the client should expect the server to issue a token in a
 * subsequent challenge.  The client must pass the token to another call
 * to GSS_Init_sec_context, repeating the actions in this paragraph.
 */
static int client_establish_context(gssldap_client_state_t state,
				    char *service_name)
{
    gss_buffer_desc send_tok, recv_tok, *token_ptr;
    gss_name_t target_name;
    OM_uint32 maj_stat, min_stat;
    gss_OID oid = GSS_C_NULL_OID;
    OM_uint32 ret_flags;

    TRACE("==> client_establish_context");

    send_tok.value = service_name;
    send_tok.length = strlen(send_tok.value) + 1;

#ifdef HACK_SERVICE_NAME
    oid = __service ? gss_nt_krb5_name : gss_nt_service_name;
    maj_stat = gss_import_name(&min_stat, &send_tok, oid, &target_name);
    oid = GSS_C_NULL_OID;
#else
    maj_stat = gss_import_name(&min_stat,
			       &send_tok,
#ifdef HEIMDAL
			       GSS_C_NT_HOSTBASED_SERVICE,
#elif SEAM
	 (gss_OID) GSS_C_NT_HOSTBASED_SERVICE,
#else
			       (gss_OID) gss_nt_service_name,
#endif				/* HEIMDAL */
			       &target_name);
#endif
    if (maj_stat != GSS_S_COMPLETE) {
	LOG_STATUS("parsing name", maj_stat, min_stat);
	TRACE("<== client_establish_context");
	return -1;
    }
     token_ptr = GSS_C_NO_BUFFER;
    state->context = GSS_C_NO_CONTEXT;

    do {
	maj_stat = gss_init_sec_context(&min_stat,
					GSS_C_NO_CREDENTIAL,
					&state->context,
					target_name,
					oid,
				   GSS_C_MUTUAL_FLAG | GSS_C_REPLAY_FLAG,
					0,
					NULL,
					token_ptr,
					NULL,
					&send_tok,
					&ret_flags,
					NULL);

	if (token_ptr != GSS_C_NO_BUFFER)
	    free(recv_tok.value); //(void) gss_release_buffer(&min_stat, &recv_tok);

	if (maj_stat != GSS_S_COMPLETE && maj_stat != GSS_S_CONTINUE_NEEDED) {
	    LOG_STATUS("initializing context", maj_stat, min_stat);
	    (void) gss_release_name(&min_stat, &target_name);
	    TRACE("<== client_establish_context");
	    return -1;
	}
#ifdef GSSSASL_DEBUG
	if (debug_) {
	    fprintf(stderr, "Sending init_sec_context token (size=%d)...\n",
		    send_tok.length);
	    fflush(stderr);
	    print_token(&send_tok);
	}
#endif				/* GSSSASL_DEBUG */	
	/* according to 7.2.1 of RFC 2222, we can/should send empty tokens */
	if (send_token(state, &send_tok) < 0) {
	    TRACE("Send_token failed");
	    (void) gss_release_buffer(&min_stat, &send_tok);
	    (void) gss_release_name(&min_stat, &target_name);
	    TRACE("<== client_establish_context");
	    return -1;
	}
	/* this is allocated by the GSS-API */
	(void) gss_release_buffer(&min_stat, &send_tok);

	if (maj_stat == GSS_S_CONTINUE_NEEDED) {
	    TRACE("continue needed...");
	    /* recv_token allocates memory for us */
	    if (recv_token(state, &recv_tok) < 0) {
		TRACE("recv_token failed");
		(void) gss_release_name(&min_stat, &target_name);
		TRACE("<== client_establish_context");
		return -1;
	    }
#ifdef GSSSASL_DEBUG
	    if (debug_) {
		fprintf(stderr, "Received token (size=%d)...\n",
			recv_tok.length);
		fflush(stderr);
		print_token(&recv_tok);
	    }
#endif				/* GSSSASL_DEBUG */
	    token_ptr = &recv_tok;
	}
    } while (maj_stat == GSS_S_CONTINUE_NEEDED);

    (void) gss_release_name(&min_stat, &target_name);

    TRACE("<== client_establish_context");

    return 0;
}

/*
 * When GSS_Init_sec_context returns GSS_S_COMPLETE, the client takes
 * the following actions: If the last call to GSS_Init_sec_context
 * returned an output_token, then the client responds with the
 * output_token, otherwise the client responds with no data.  The client
 * should then expect the server to issue a token in a subsequent
 * challenge.  The client passes this token to GSS_Unwrap and interprets
 * the first octet of resulting cleartext as a bit-mask specifying the
 * security layers supported by the server and the second through fourth
 * octets as the maximum size output_message to send to the server.  The
 * client then constructs data, with the first octet containing the
 * bit-mask specifying the selected security layer, the second through
 * fourth octets containing in network byte order the maximum size
 * output_message the client is able to receive, and the remaining
 * octets containing the authorization identity.  The client passes the
 * data to GSS_Wrap with conf_flag set to FALSE, and responds with the
 * generated output_message.  The client can then consider the server
 * authenticated.
 */
static int negotiate_security_options(gssldap_client_state_t state, int layer)
	{
	OM_uint32 maj_stat, min_stat;
	gss_buffer_desc recv_tok, send_tok;
	int rc;
	size_t mask_length;
	gsssasl_security_negotiation_t send_mask, recv_mask;
	char *identity;

	TRACE("==> negotiate_security_options");

	if ((rc = recv_token(state, &recv_tok)) < 0) {
	TRACE("<== negotiate_security_options (recv_token failed)");
	return rc;
	}
	#ifdef GSSSASL_DEBUG
	if (debug_) {
	fprintf(stderr, "Received token (size=%d)...\n",
		recv_tok.length);
	fflush(stderr);
	print_token(&recv_tok);
	}
	#endif				/* GSSSASL_DEBUG */
	print_token(&recv_tok); //added
	maj_stat = gsssasl_unpack_security_token(&min_stat,
							state->context,
							&recv_tok,
							&recv_mask,
							&mask_length);

	if (maj_stat != GSS_S_COMPLETE) {
	LOG_STATUS("unpacking security negotiation token",
			maj_stat,
			min_stat);
	TRACE("<== negotiate_security_options (unpack failed)");
	return -1;
	}
	#ifdef GSSSASL_DEBUG
	if (debug_) {
	fprintf(stderr, "Received security token level %d size %d\n",
		recv_mask->security_layer,
		recv_mask->token_size);
	}
	#endif				/* GSSSASL_DEBUG */

	if ((~recv_mask->security_layer) & layer) {
		free(recv_mask);
		TRACE("<== negotiate_security_options (unsupported security layer)");
		return -1;
	}

	/*
	** send the DN here because it gets wrapped
	*/
	mask_length = sizeof(recv_mask); //+ GSSAPI_LDAP_DN_PREFIX_LEN + strlen(state->binddn);//vikas
	send_mask = NSLDAPI_MALLOC(mask_length); 	
	if (send_mask == NULL) {
		free(recv_mask);
		TRACE("<== negotiate_security_options (malloc failed)");
		return -1;
	}
	send_mask->security_layer = layer;
	send_mask->token_size = recv_mask->token_size;
	/*
	identity = (char *)send_mask + sizeof(gsssasl_security_negotiation_desc);

	memcpy(identity,
		GSSAPI_LDAP_DN_PREFIX,
		GSSAPI_LDAP_DN_PREFIX_LEN);
	memcpy(identity + GSSAPI_LDAP_DN_PREFIX_LEN,
		state->binddn,
		mask_length - sizeof(recv_mask) - GSSAPI_LDAP_DN_PREFIX_LEN);
	*/
	/* allocated by gsssasl_unpack_security_token */
	free(recv_mask);

	#ifdef GSSSASL_DEBUG
	if (debug_) {
		fprintf(stderr, "Sending security token level %d size %d\n",
		send_mask->security_layer,
		send_mask->token_size);
	}
	#endif				/* GSSSASL_DEBUG */

	maj_stat = gsssasl_pack_security_token(&min_stat,
						state->context,
						send_mask,
						mask_length,
						&send_tok);

	if (maj_stat != GSS_S_COMPLETE) {
		LOG_STATUS("packing security negotiation token",
				maj_stat,
			min_stat);
		/* allocated locally */
		NSLDAPI_FREE(send_mask);  	
		TRACE("<== negotiate_security_options (pack failed)");
		return -1;
	}
	if ((rc = send_token(state, &send_tok)) < 0) {
		/* allocated locally */
		NSLDAPI_FREE(send_mask); 	
		TRACE("<== negotiate_security_options (send_token failed)");
		return rc;
	}
	rc = parse_bind_result(state);

	if (rc == 0) {
		security_context = state->context;
		security_layer = layer;
		security_token_size = send_mask->token_size;
	}else{
		fprintf(stdout, "rc !=0 after parse_bind_result\n");
	}
	/* allocated locally */
	NSLDAPI_FREE(send_mask); 
	TRACE("<== negotiate_security_options");

	return rc;
}
LDAP_API(int) LDAP_C LDAP_CALLBACK ldap_gssapi_read(LBER_SOCKET sock, void *data, int len)
{
    OM_uint32 maj_stat, min_stat, rc;
    gss_buffer_desc recv_tok, wrapped_tok;
    OM_uint32 pdulen;

    if (security_layer & (GSSSASL_INTEGRITY_PROTECTION | GSSSASL_PRIVACY_PROTECTION)) {
	if (read(sock, &pdulen, sizeof(pdulen)) != sizeof(pdulen))
	    return (0);
	wrapped_tok.length = ntohl(pdulen);

#ifdef GSSSASL_DEBUG
	if (debug_) {
	    fprintf(stderr, "Reading data of %d octets\n",
		    wrapped_tok.length);
	}
#endif				// GSSSASL_DEBUG 
	if (wrapped_tok.length > security_token_size)
	    return (0);

	wrapped_tok.value = malloc(wrapped_tok.length);
	if (wrapped_tok.value == NULL)
	    return (0);
	if (read(sock, wrapped_tok.value, wrapped_tok.length) != wrapped_tok.length) {
	    gss_release_buffer(&rc, &wrapped_tok);
	    return (0);
	}
	maj_stat = gss_unwrap(&min_stat,
			      security_context,
			      &wrapped_tok,
			      &recv_tok,
			      &rc,
			      (gss_qop_t *) NULL);
	if (maj_stat != GSS_S_COMPLETE) {
	    gss_release_buffer(&rc, &wrapped_tok);
	    return maj_stat;
	}
#ifdef GSSSASL_DEBUG
	if (debug_) {
	    fprintf(stderr, "Got %d bytes of %s data\n",
		    recv_tok.length, rc ? "private" : "signed");
	}
#endif				// GSSSASL_DEBUG 

	if (recv_tok.length > len) {
	    gss_release_buffer(&rc, &recv_tok);
	    gss_release_buffer(&rc, &wrapped_tok);
	    return GSS_S_FAILURE;
	}
	memcpy(data, recv_tok.value, recv_tok.length);
	pdulen = recv_tok.length;
	gss_release_buffer(&rc, &recv_tok);
	gss_release_buffer(&rc, &wrapped_tok);

	return pdulen;
    } else
	return read(sock, data, len);
}
LDAP_API(int) LDAP_C LDAP_CALLBACK ldap_gssapi_write(LBER_SOCKET sock, const void *data, int len)
{
    OM_uint32 maj_stat, min_stat, rc;
    gss_buffer_desc send_tok, wrapped_tok;
    OM_uint32 pdulen;
    struct iovec iov[2];

    if (security_layer & (GSSSASL_INTEGRITY_PROTECTION | GSSSASL_PRIVACY_PROTECTION)) {
	send_tok.length = len;
	send_tok.value = (void *) data;

#ifdef GSSSASL_DEBUG
	if (debug_) {
	    fprintf(stderr, "Sending %d bytes of data\n",
		    len);
	}
#endif				 //GSSSASL_DEBUG 

	maj_stat = gss_wrap(&min_stat,
			    security_context,
		   (security_layer & GSSSASL_PRIVACY_PROTECTION) ? 1 : 0,
			    GSS_C_QOP_DEFAULT,
			    &send_tok,
			    &rc,
			    &wrapped_tok);

	if (maj_stat != GSS_S_COMPLETE)
	    return maj_stat;

#ifdef GSSSASL_DEBUG
	if (debug_) {
	    fprintf(stderr, "Sent %d bytes of %s data\n",
		    wrapped_tok.length, rc ? "private" : "signed");
	}
#endif				// GSSSASL_DEBUG 

	pdulen = htonl(wrapped_tok.length);

	iov[0].iov_base = (void *)&pdulen;
	iov[0].iov_len = sizeof(pdulen);
	iov[1].iov_base = wrapped_tok.value;
	iov[1].iov_len = wrapped_tok.length;

	rc = send(sock, iov, 2,MSG_DONTROUTE);

//	gss_release_buffer(&maj_stat, &send_tok); 
	gss_release_buffer(&maj_stat, &wrapped_tok);

	if (rc == (wrapped_tok.length + sizeof(pdulen)))
	    rc = len;

	return rc;
    } else
	return write(sock, data, len);
}


#ifdef GSSSASL_DEBUG
/*
 * Temporary function to enable debugging
 */
LDAP_API(int)
LDAP_CALL ldap_gssapi_debug(int on)
{
    int old = debug_;
    debug_ = on;
    return old;
}

#endif				/* GSSSASL_DEBUG */

/*
 * Public function for doing a GSS-API SASL bind
 */
LDAP_API(int)
LDAP_CALL ldap_gssapi_bind(LDAP * ld, const char *who, int layer)
{
    ///////////////////
	gssldap_client_state_desc state;
	gss_buffer_desc sname, tname;
		gss_name_t src_name, targ_name;
		
	OM_uint32 context_flags;
	OM_uint32 lifetime, maj_stat;
	int is_open, is_local;
	gss_OID mechanism, name_type;
	////////////////////////////
    char *service_name;
    OM_uint32 min_stat;
    int rc;

    /* not sure about this defhost stuff. Looks a bit dangerous. */
    if (!NSLDAPI_VALID_LDAP_POINTER(ld) || ld->ld_defhost == NULL) {
	return -1;
    }
    service_name = NSLDAPI_MALLOC(sizeof(GSSAPI_LDAP_SERVICE_NAME "@") + strlen(ld->ld_defhost));
    if (service_name == NULL) {
	return -1;
    }
#ifdef HACK_SERVICE_NAME
    if (__service) {
	fprintf(stderr, "WARNING -- setting service name to %s\n", __service);
	free(service_name);
	service_name = strdup(__service);
    } else {
#endif				/* HACK_SERVICE_NAME */
	strcpy(service_name, GSSAPI_LDAP_SERVICE_NAME "@");
	strcat(service_name, ld->ld_defhost);
#ifdef HACK_SERVICE_NAME
    }
#endif				/* HACK_SERVICE_NAME */

#ifdef GSSSASL_DEBUG
    if (debug_) {
		fprintf(stderr, "LDAP service name: %s\n", service_name);
		fflush(stderr);
    }
#endif				/* GSSSASL_DEBUG */

    state.msgid = -1;
    state.ld = ld;
    fprintf(stdout,"\n\n\nwho=%s",who);
    state.binddn = who;
    state.context = GSS_C_NO_CONTEXT;
    state.rc = LDAP_OPERATIONS_ERROR;

    rc = client_establish_context(&state, service_name);
    if (rc == 0) {
	rc = negotiate_security_options(&state, layer);
   }
fprintf(stdout, "\nafter negotiate_security_options %d\n ",rc);
    if (rc == 0) {
	fprintf(stdout, "goes inside \n");
#ifdef GSSSASL_DEBUG
//#ifdef 0
	/* For Debugging only */
	//gss_buffer_desc sname, tname;


	maj_stat = gss_inquire_context(&min_stat,
				       state.context,
				       &src_name,
				       &targ_name,
				       &lifetime,
				       &mechanism,
				       &context_flags,
				       &is_local,
				       &is_open);
	if (maj_stat != GSS_S_COMPLETE) {
	    LOG_STATUS("inquiring context", maj_stat, min_stat);
	    return LDAP_OPERATIONS_ERROR;
	}
	maj_stat = gss_display_name(&min_stat,
				    src_name,
				    &sname,
				    &name_type);
	if (maj_stat != GSS_S_COMPLETE) {
	    LOG_STATUS("displaying source name", maj_stat, min_stat);
	    return LDAP_OPERATIONS_ERROR;
	}
	maj_stat = gss_display_name(&min_stat,
				    targ_name,
				    &tname,
				    (gss_OID *) NULL);
	if (maj_stat != GSS_S_COMPLETE) {
	    LOG_STATUS("displaying target name", maj_stat, min_stat);
	    return LDAP_OPERATIONS_ERROR;
	}
	if (debug_) {
	    fprintf(stderr, "\"%.*s\" to \"%.*s\", lifetime %d, flags %x, %s, %s\n",
		    (int) sname.length, (char *) sname.value,
		    (int) tname.length, (char *) tname.value, lifetime,
		    context_flags,
		 (is_local) ? "locally initiated" : "remotely initiated",
		    (is_open) ? "open" : "closed");
	    fflush(stderr);
	}
	(void) gss_release_name(&min_stat, &src_name);
	(void) gss_release_name(&min_stat, &targ_name);
	(void) gss_release_buffer(&min_stat, &sname);
	(void) gss_release_buffer(&min_stat, &tname);
#endif				/* GSSSASL_DEBUG */
	state.rc = LDAP_SUCCESS;
    }
#if 0
    if (state.context != GSS_C_NO_CONTEXT) {
	(void) gss_delete_sec_context(&min_stat, &state.context, GSS_C_NO_BUFFER);
    }
#endif
    if (state.rc == LDAP_SUCCESS) {
	/*	struct ldap_io_fns iofns;

		memset(&iofns, 0, sizeof(iofns));
		iofns.liof_read = ldap_gssapi_read;
		iofns.liof_write = ldap_gssapi_write;

		state.rc = ldap_set_option(ld, LDAP_OPT_IO_FN_PTRS, &iofns);*/
		//struct ldap_io_fns iofns;
		struct ldap_x_ext_io_fns iofns;
		memset(&iofns, 0, sizeof(iofns));
		iofns.lextiof_read = ldap_gssapi_read;
		iofns.lextiof_write = ldap_gssapi_write;
		state.rc = ldap_set_option(ld, LDAP_X_OPT_EXTIO_FN_PTRS, (void *)&iofns /* &iofns*/)  ;
	    
	}
    NSLDAPI_FREE(service_name);	
    LDAP_SET_LDERRNO(ld, state.rc, NULL, NULL);

    return state.rc;
}

